import json
from typing import Optional
from fastapi import HTTPException
from lmos_openai_types import (
    ChatCompletionMessageToolCall,
    ChatCompletionRequestMessage,
    CreateChatCompletionRequest,
    CreateChatCompletionStreamResponse,
    Function1,
)
from .utils import call_tool, chat_completion_add_tools
from mcp_bridge.models import SSEData
from .genericHttpxClient import client
from mcp_bridge.mcp_clients.McpClientManager import ClientManager
from mcp_bridge.tool_mappers import mcp2openai
from loguru import logger
from httpx_sse import aconnect_sse

from sse_starlette.sse import EventSourceResponse, ServerSentEvent


async def streaming_chat_completions(request: CreateChatCompletionRequest):
    # raise NotImplementedError("Streaming Chat Completion is not supported")

    try:
        return EventSourceResponse(
            content=chat_completions(request),
            media_type="text/event-stream",
            headers={"Cache-Control": "no-cache"},
        )

    except Exception as e:
        logger.error(e)


async def chat_completions(request: CreateChatCompletionRequest):
    """performs a chat completion using the inference server"""

    request.stream = True

    request = await chat_completion_add_tools(request)

    fully_done = False
    while not fully_done:
        # json_data = request.model_dump_json(
        #     exclude_defaults=True, exclude_none=True, exclude_unset=True
        # )

        json_data = json.dumps(request.model_dump(
            exclude_defaults=True, exclude_none=True, exclude_unset=True
        ))

        # logger.debug(json_data)

        last: Optional[CreateChatCompletionStreamResponse] = None  # last message

        tool_call_name: str = ""
        tool_call_json: str = ""
        should_forward: bool = True
        response_content: str = ""
        tool_call_id: str = ""

        async with aconnect_sse(
            client, "post", "/chat/completions", content=json_data
        ) as event_source:
            
            # check if the content type is correct because the aiter_sse method
            # will raise an exception if the content type is not correct
            if "Content-Type" in event_source.response.headers:
                content_type = event_source.response.headers["Content-Type"]
                if "text/event-stream" not in content_type:
                    logger.error(f"Unexpected Content-Type: {content_type}")
                    error_data = await event_source.response.aread()
                    logger.error(f"Request URL: {event_source.response.url}")
                    logger.error(f"Request Data: {json_data}")
                    logger.error(f"Response Status: {event_source.response.status_code}")
                    logger.error(f"Response Data: {error_data.decode(event_source.response.encoding or 'utf-8')}")
                    raise HTTPException(status_code=500, detail="Unexpected Content-Type")

            # iterate over the SSE stream
            async for sse in event_source.aiter_sse():
                event = sse.event
                data = sse.data
                id = sse.id
                retry = sse.retry

                logger.debug(
                    f"event: {event},\ndata: {data},\nid: {id},\nretry: {retry}"
                )

                # handle if the SSE stream is done
                if data == "[DONE]":
                    logger.debug("inference serverstream done")
                    break

                # for some reason openrouter uses uppercase for finish_reason
                try:
                    data['choices'][0]['finish_reason'] = data['choices'][0]['finish_reason'].lower() # type: ignore
                except Exception as e:
                    logger.debug(f"failed to lowercase finish_reason: {e}")

                try:
                    parsed_data = CreateChatCompletionStreamResponse.model_validate_json(
                        data
                    )
                except Exception as e:
                    logger.debug(data)
                    raise e

                # add the delta to the response content
                content = parsed_data.choices[0].delta.content
                content = content if content is not None else ""
                response_content += content

                # handle stop reasons
                if parsed_data.choices[0].finish_reason is not None:
                    if parsed_data.choices[0].finish_reason.value in [
                        "stop",
                        "length",
                    ]:
                        fully_done = True
                    else:
                        should_forward = False

                # this manages the incoming tool call schema
                # most of this is assertions to please mypy
                if parsed_data.choices[0].delta.tool_calls is not None:
                    should_forward = False
                    assert (
                        parsed_data.choices[0].delta.tool_calls[0].function is not None
                    )

                    name = parsed_data.choices[0].delta.tool_calls[0].function.name
                    name = name if name is not None else ""
                    tool_call_name = name if tool_call_name == "" else tool_call_name

                    call_id = parsed_data.choices[0].delta.tool_calls[0].id
                    call_id = call_id if call_id is not None else ""
                    tool_call_id = id if tool_call_id == "" else tool_call_id

                    arg = parsed_data.choices[0].delta.tool_calls[0].function.arguments
                    tool_call_json += arg if arg is not None else ""

                # forward SSE messages to the client
                logger.debug(f"{should_forward=}")
                if should_forward:
                    # we do not want to forward tool call json to the client
                    logger.debug("forwarding message")
                    yield SSEData.model_validate_json(sse.data).model_dump_json()

                # save the last message
                last = parsed_data

        # ideally we should check this properly
        assert last is not None
        assert last.choices[0].finish_reason is not None

        if last.choices[0].finish_reason.value in ["stop", "length"]:
            logger.debug("no tool calls found")
            fully_done = True
            continue

        logger.debug("tool calls found")
        logger.debug(
            f"{tool_call_name=} {tool_call_json=}"
        )  # this should not be error but its easier to debug

        # add received message to the history
        msg = ChatCompletionRequestMessage(
            role="assistant",
            content=response_content,
            tool_calls=[
                ChatCompletionMessageToolCall(
                    id=tool_call_id,
                    type="function",
                    function=Function1(name=tool_call_name, arguments=tool_call_json),
                )
            ],
        )  # type: ignore
        request.messages.append(msg)

        #### MOST OF THIS IS COPY PASTED FROM CHAT_COMPLETIONS
        # FIXME: this can probably be done in parallel using asyncio gather
        tool_call_result = await call_tool(tool_call_name, tool_call_json)
        if tool_call_result is None:
            continue

        logger.debug(
            f"tool call result for {tool_call_name}: {tool_call_result.model_dump()}"
        )

        logger.debug(f"tool call result content: {tool_call_result.content}")

        tools_content = [
            {"type": "text", "text": part.text}
            for part in filter(lambda x: x.type == "text", tool_call_result.content)
        ]
        if len(tools_content) == 0:
            tools_content = [{"type": "text", "text": "the tool call result is empty"}]
        request.messages.append(
            ChatCompletionRequestMessage.model_validate(
                {
                    "role": "tool",
                    "content": tools_content,
                    "tool_call_id": tool_call_id,
                }
            )
        )

        logger.debug("sending next iteration of chat completion request")

    # when done, send the final event
    logger.debug("sending final event")
    yield ServerSentEvent(event="message", data="[DONE]", id=None, retry=None)



from fastapi import APIRouter

from lmos_openai_types import CreateChatCompletionRequest, CreateCompletionRequest

from mcp_bridge.openai_clients import (
    client,
    completions,
    chat_completions,
    streaming_chat_completions,
)

from mcp_bridge.openapi_tags import Tag

router = APIRouter(prefix="/v1", tags=[Tag.openai])


@router.post("/completions")
async def openai_completions(request: CreateCompletionRequest):
    """Completions endpoint"""
    if request.stream:
        raise NotImplementedError("Streaming Completion is not supported")
    else:
        return await completions(request)


@router.post("/chat/completions")
async def openai_chat_completions(request: CreateChatCompletionRequest):
    """Chat Completions endpoint"""
    if request.stream:
        return await streaming_chat_completions(request)
    else:
        return await chat_completions(request)


@router.get("/models")
async def models():
    """List models"""
    response = await client.get("/models")
    return response.json()



from pydantic_settings import BaseSettings, SettingsConfigDict
from pydantic import Field, Json
from typing import Optional

__all__ = ["initial_settings"]


class InitialSettings(BaseSettings):
    file: Optional[str] = Field("config.json")
    http_url: Optional[str] = Field(None)
    json: Optional[Json] = Field(None)  # allow for raw config to be passed as env var

    load_config: bool = Field(
        True, include_in_schema=False
    )  # this can be used to disable loading the config

    model_config = SettingsConfigDict(
        env_prefix="MCP_BRIDGE__CONFIG__",
        env_file=".env",
        env_file_encoding="utf-8",
        env_nested_delimiter="__",
    )


# This will load the InitialSettings from environment variables
initial_settings = InitialSettings()



import json
import httpx
from typing import Any
from loguru import logger


def load_config(url: str) -> dict[str, Any]:
    try:
        resp = httpx.get(str(url))
        return resp.json()

    except httpx.ConnectError:
        logger.error(f"could not connect to {httpx.URL(url).host}")

    except json.JSONDecodeError:
        logger.error(f"failed to parse json from {httpx.URL(url)}")

    return {}



from typing import Union
from mcp_bridge.config import config
from mcp import McpError, StdioServerParameters
from loguru import logger

from .StdioClient import StdioClient
from .SseClient import SseClient
from .DockerClient import DockerClient
from mcp_bridge.config.final import DockerMCPServer, SSEMCPServer

client_types = Union[StdioClient, SseClient, DockerClient]


class MCPClientManager:
    clients: dict[str, client_types] = {}

    async def initialize(self):
        """Initialize the MCP Client Manager and start all clients"""

        logger.log("DEBUG", "Initializing MCP Client Manager")

        for server_name, server_config in config.mcp_servers.items():
            self.clients[server_name] = await self.construct_client(
                server_name, server_config
            )

    async def construct_client(self, name, server_config) -> client_types:
        logger.log("DEBUG", f"Constructing client for {server_config}")

        if isinstance(server_config, StdioServerParameters):
            client = StdioClient(name, server_config)
            await client.start()
            return client

        if isinstance(server_config, SSEMCPServer):
            # TODO: implement sse client
            client = SseClient(name, server_config)  # type: ignore
            await client.start()
            return client
        
        if isinstance(server_config, DockerMCPServer):
            client = DockerClient(name, server_config)
            await client.start()
            return client

        raise NotImplementedError("Client Type not supported")

    def get_client(self, server_name: str):
        return self.clients[server_name]

    def get_clients(self):
        return list(self.clients.items())

    async def get_client_from_tool(self, tool: str):
        for name, client in self.get_clients():
            
            # client cannot have tools if it is not connected
            if not client.session:
                continue

            try:
                list_tools = await client.session.list_tools()
                for client_tool in list_tools.tools:
                    if client_tool.name == tool:
                        return client
            except McpError:
                continue

    async def get_client_from_prompt(self, prompt: str):
        for name, client in self.get_clients():

            # client cannot have prompts if it is not connected
            if not client.session:
                continue

            try:
                list_prompts = await client.session.list_prompts()
                for client_prompt in list_prompts.prompts:
                    if client_prompt.name == prompt:
                        return client
            except McpError:
                continue


ClientManager = MCPClientManager()



from fastapi import APIRouter, HTTPException
from mcp_bridge.mcp_clients.McpClientManager import ClientManager
from mcp.types import ListResourcesResult

router = APIRouter(prefix="/resources")


@router.get("")
async def get_resources() -> dict[str, ListResourcesResult]:
    """Get all resources from all MCP clients"""

    resources = {}

    for name, client in ClientManager.get_clients():
        resources[name] = await client.list_resources()

    return resources



from lmos_openai_types import CreateCompletionRequest
from .genericHttpxClient import client


async def completions(request: CreateCompletionRequest) -> dict:
    """performs a completion using the inference server"""

    response = await client.post(
        "/completions",
        json=request.model_dump(
            exclude_defaults=True, exclude_none=True, exclude_unset=True
        ),
    )
    return response.json()



from fastapi import APIRouter
from .sse import router as sse_router
from mcp_bridge.openapi_tags import Tag

__all__ = ["router"]

router = APIRouter(prefix="/mcp-server", tags=[Tag.mcp_server])
router.include_router(sse_router)



import math

from mcp.types import ModelPreferences

from mcp_bridge.config import config

def euclidean_distance(point1, point2):
    """
    Calculates the Euclidean distance between two points, ignoring None values.
    """
    valid_dimensions = [(p1, p2) for p1, p2 in zip(point1, point2) if p1 is not None and p2 is not None]

    if not valid_dimensions:  # No valid dimensions to compare
        return float('inf')
    
    return math.sqrt(sum((p1 - p2) ** 2 for p1, p2 in valid_dimensions))

def find_best_model(preferences: ModelPreferences):
    distance = math.inf
    preffered_model = None
    preference_points = (preferences.intelligencePriority, preferences.speedPriority, preferences.costPriority)

    if preference_points == (None, None, None):
        return config.sampling.models[0]

    for model in config.sampling.models:
        model_points = (model.intelligence, model.speed, model.cost)
        model_distance = euclidean_distance(model_points, preference_points)
        if model_distance < distance:
            distance = model_distance
            preffered_model = model
    
    if preffered_model is None:
        preffered_model = config.sampling.models[0]
    
    return preffered_model


from typing import Optional, List
from pydantic import BaseModel


class Delta(BaseModel):
    role: Optional[str] = None
    content: Optional[str] = None


class Choice(BaseModel):
    index: int
    delta: Delta
    logprobs: Optional[dict] = None
    finish_reason: Optional[str] = None


class SSEData(BaseModel):
    id: str
    object: str
    created: int
    model: str
    choices: List[Choice]



from pydantic import BaseModel, Field


class McpServerStatus(BaseModel):
    name: str = Field(..., description="Name of the MCP server")
    online: bool = Field(..., description="Whether the server is online")
    enabled: bool = Field(True, description="Whether the server is enabled")



from loguru import logger
from mcp import SamplingMessage
import mcp.types as types
from lmos_openai_types import CreateChatCompletionResponse
from mcp.types import CreateMessageRequestParams, CreateMessageResult

from mcp_bridge.config import config
from mcp_bridge.openai_clients.genericHttpxClient import client
from mcp_bridge.sampling.modelSelector import find_best_model

def make_message(x: SamplingMessage):
    if x.content.type == "text":
        return {
            "role": x.role,
            "content": [{
                "type": "text",
                "text": x.content.text,
            }]
        }
    if x.content.type == "image":
        return {
            "role": x.role,
            "content": [{
                "type": "image",
                "image_url": x.content.data,
            }]
        }

async def handle_sampling_message(
    message: CreateMessageRequestParams,
) -> CreateMessageResult:
    """perform sampling"""

    logger.debug(f"sampling message: {message.modelPreferences}")
    
    # select model
    model = config.sampling.models[0]
    if message.modelPreferences is not None:
        model = find_best_model(message.modelPreferences)

    logger.debug(f"selected model: {model.model}")

    logger.debug("sending sampling request to endpoint")
    # request = CreateChatCompletionRequest(model=model.model, messages=message.messages, stream=False)  # type: ignore
    request = {
        "model": model.model,
        "messages": [make_message(x) for x in message.messages],
        "stream": False,
    }

    logger.debug(f"request: {request}")

    logger.debug(request)

    resp = await client.post(
        "/chat/completions",
        json=request,
        timeout=config.sampling.timeout,
    )

    logger.debug("parsing json")
    text = resp.text
    logger.debug(text)

    response = CreateChatCompletionResponse.model_validate_json(text)

    logger.debug("sampling request received from endpoint")

    assert response.choices is not None
    assert len(response.choices) > 0
    assert response.choices[0].message is not None
    assert response.choices[0].message.content is not None

    return types.CreateMessageResult(
        role="assistant",
        content=types.TextContent(
            type="text",
            text=response.choices[0].message.content,
        ),
        model=model.model,
        stopReason=response.choices[0].finish_reason,
    )



from fastapi import APIRouter, HTTPException
from mcp.types import ListPromptsResult, ListToolsResult, ListResourcesResult
from mcp_bridge.models.mcpServerStatus import McpServerStatus
from mcp_bridge.mcp_clients.McpClientManager import ClientManager

router = APIRouter(prefix="/servers")


@router.get("/{server_name}/prompts")
async def get_server_prompts(server_name: str) -> ListPromptsResult:
    """Get all prompts from a specific MCP server"""

    client = ClientManager.get_client(server_name)
    if not client:
        raise HTTPException(status_code=404, detail=f"Server '{server_name}' not found")

    return await client.list_prompts()


@router.get("/{server_name}/tools")
async def get_server_tools(server_name: str) -> ListToolsResult:
    """Get all tools from a specific MCP server"""

    client = ClientManager.get_client(server_name)
    if not client:
        raise HTTPException(status_code=404, detail=f"Server '{server_name}' not found")

    return await client.list_tools()


@router.get("/{server_name}/resources")
async def get_server_resources(server_name: str) -> ListResourcesResult:
    """Get all resources from a specific MCP server"""

    client = ClientManager.get_client(server_name)
    if not client:
        raise HTTPException(status_code=404, detail=f"Server '{server_name}' not found")

    return await client.list_resources()


@router.get("/{server_name}/status")
async def get_server_status(server_name: str) -> McpServerStatus:
    """Get the status of a specific MCP server"""

    client = ClientManager.get_client(server_name)
    if not client:
        raise HTTPException(status_code=404, detail=f"Server '{server_name}' not found")

    return await client.status()



import json
from typing import Any
from loguru import logger


def load_config(file: str) -> dict[str, Any]:
    try:
        with open(file, "r") as f:
            return json.load(f)

    except FileNotFoundError:
        logger.warning(f'the "{file}" file was not found')

    except Exception:
        logger.error(f'there was an error reading the "{file}" file')

    return {}



import asyncio
from abc import ABC, abstractmethod
from typing import Any, Optional
from fastapi import HTTPException
from mcp import McpError
from mcp.types import (
    CallToolResult,
    ListToolsResult,
    TextContent,
    ListResourcesResult,
    ListPromptsResult,
    GetPromptResult,
    TextResourceContents,
    BlobResourceContents,
)
from loguru import logger
from pydantic import AnyUrl
from mcp_bridge.mcp_clients.session import McpClientSession
from mcp_bridge.models.mcpServerStatus import McpServerStatus


class GenericMcpClient(ABC):
    name: str
    config: Any
    client: Any
    session: McpClientSession | None = None

    def __init__(self, name: str) -> None:
        super().__init__()
        self.session = None
        self.name = name

        logger.debug(f"initializing client class for {name}")

    @abstractmethod
    async def _maintain_session(self):
        pass

    async def _session_maintainer(self):
        while True:
            try:
                await self._maintain_session()
            except FileNotFoundError as e:
                logger.error(f"failed to maintain session for {self.name}: file {e.filename} not found.")
            except Exception as e:
                logger.error(f"failed to maintain session for {self.name}: {type(e)} {e.args}")

            logger.debug(f"restarting session for {self.name}")
            await asyncio.sleep(0.5)

    async def start(self):
        asyncio.create_task(self._session_maintainer())

    async def call_tool(
        self, name: str, arguments: dict, timeout: Optional[int] = None
    ) -> CallToolResult:
        await self._wait_for_session()

        try:
            async with asyncio.timeout(timeout):
                return await self.session.call_tool(
                    name=name,
                    arguments=arguments,
                )

        except asyncio.TimeoutError:
            logger.error(f"timed out calling tool: {name}")
            return CallToolResult(
                content=[
                    TextContent(type="text", text=f"Timeout Error calling {name}")
                ],
                isError=True,
            )

        except McpError as e:
            logger.error(f"error calling {name}: {e}")
            return CallToolResult(
                content=[TextContent(type="text", text=f"Error calling {name}: {e}")],
                isError=True,
            )

    async def get_prompt(
        self, prompt: str, arguments: dict[str, str]
    ) -> GetPromptResult | None:
        await self._wait_for_session()

        try:
            return await self.session.get_prompt(prompt, arguments)
        except Exception as e:
            logger.error(f"error evaluating prompt: {e}")

        return None

    async def read_resource(
        self, uri: AnyUrl
    ) -> list[TextResourceContents | BlobResourceContents]:
        await self._wait_for_session()
        try:
            resource = await self.session.read_resource(uri)
            return resource.contents
        except Exception as e:
            logger.error(f"error reading resource: {e}")
            return []

    async def list_tools(self) -> ListToolsResult:
        # if session is None, then the client is not running
        # wait to see if it restarts
        await self._wait_for_session()

        try:
            return await self.session.list_tools()
        except Exception as e:
            logger.error(f"error listing tools: {e}")
            return ListToolsResult(tools=[])

    async def list_resources(self) -> ListResourcesResult:
        await self._wait_for_session()
        try:
            return await self.session.list_resources()
        except Exception as e:
            logger.error(f"error listing resources: {e}")
            return ListResourcesResult(resources=[])

    async def list_prompts(self) -> ListPromptsResult:
        await self._wait_for_session()
        try:
            return await self.session.list_prompts()
        except Exception as e:
            logger.error(f"error listing prompts: {e}")
            return ListPromptsResult(prompts=[])

    async def _wait_for_session(self, timeout: int = 5, http_error: bool = True):
        try:
            async with asyncio.timeout(timeout):
                while self.session is None:
                    await asyncio.sleep(1)
                    logger.debug(f"waiting for session for {self.name}")

        except asyncio.TimeoutError:
            if http_error:
                raise HTTPException(
                    status_code=500, detail=f"Could not connect to MCP server \"{self.name}\"." 
                )

            raise TimeoutError(f"Could not connect to MCP server \"{self.name}\"." )

        assert self.session is not None, "Session is None"

    async def status(self) -> McpServerStatus:
        """Get the status of the MCP server"""
        return McpServerStatus(
            name=self.name, online=self.session is not None, enabled=True
        )



from mcp import types
from mcp.server import Server, NotificationOptions
from mcp.server.models import InitializationOptions
from pydantic import AnyUrl
from mcp_bridge.mcp_clients.McpClientManager import ClientManager
from loguru import logger

__all__ = ["server", "options"]

server = Server("MCP-Bridge")

## list functions


@server.list_prompts()
async def list_prompts() -> list[types.Prompt]:
    prompts = []
    for name, client in ClientManager.get_clients():
        # if client is None, then we cannot list the prompts
        if client is None:
            logger.error(f"Client '{name}' not found")
            continue

        client_prompts = await client.list_prompts()
        prompts.extend(client_prompts.prompts)
    return prompts


@server.list_resources()
async def list_resources() -> list[types.Resource]:
    resources = []
    for name, client in ClientManager.get_clients():
        try:
            client_resources = await client.list_resources()
            resources.extend(client_resources.resources)
        except Exception as e:
            logger.error(f"Error listing resources for {name}: {e}")
    return resources


@server.list_resource_templates()
async def list_resource_templates() -> list[types.ResourceTemplate]:
    return []


@server.list_tools()
async def list_tools() -> list[types.Tool]:
    tools = []
    for name, client in ClientManager.get_clients():
        # if client is None, then we cannot list the tools
        if client is None:
            logger.error(f"Client '{name}' not found")
            continue

        client_tools = await client.list_tools()
        tools.extend(client_tools.tools)
    return tools


## get functions


@server.get_prompt()
async def get_prompt(name: str, args: dict[str, str] | None) -> types.GetPromptResult:
    client = await ClientManager.get_client_from_prompt(name)

    # if client is None, then we cannot get the prompt
    if client is None:
        raise Exception(f"Prompt '{name}' not found")

    # if args is None, then we should use an empty dict
    if args is None:
        args = {}

    result = await client.get_prompt(name, args)
    if result is None:
        raise Exception(f"Prompt '{name}' not found")

    return result


@server.read_resource()
async def handle_read_resource(uri: AnyUrl) -> str | bytes:
    for name, client in ClientManager.get_clients():
        try:
            client_resources = await client.list_resources()
            if str(uri) in map(lambda x: str(x.uri), client_resources.resources):
                response = await client.read_resource(uri)
                for resource in response:
                    if resource.mimeType == "text/plain":
                        assert isinstance(resource, types.TextResourceContents)
                        assert type(resource.text) is str
                        return resource.text

                    elif resource.mimeType == "application/octet-stream":
                        assert isinstance(resource, types.BlobResourceContents)
                        assert type(resource.blob) is bytes
                        return resource.blob

                    else:
                        raise Exception(
                            f"Unsupported resource type: {resource.mimeType}"
                        )

        except Exception as e:
            logger.error(f"Error listing resources for {name}: {e}")

    raise Exception(f"Resource '{uri}' not found")


@server.call_tool()
async def handle_call_tool(
    name: str, arguments: dict | None
) -> list[types.TextContent | types.ImageContent | types.EmbeddedResource]:
    client = await ClientManager.get_client_from_tool(name)

    # if client is None, then we cannot call the tool
    if client is None:
        raise Exception(f"Tool '{name}' not found")

    # if arguments is None, then we should use an empty dict
    if arguments is None:
        arguments = {}

    return (await client.call_tool(name, arguments)).content


# options

options = InitializationOptions(
    server_name="MCP-Bridge",
    server_version="0.2.0",
    capabilities=server.get_capabilities(
        notification_options=NotificationOptions(),
        experimental_capabilities={},
    ),
)



from pydantic import BaseModel, Field
from typing import Literal, Optional
import datetime


class UnhealthyEvent(BaseModel):
    """Represents an unhealthy event"""

    name: str = Field(..., description="Name of the event")
    severity: Literal["error", "warning"] = Field(
        ..., description="Severity of the event"
    )
    traceback: Optional[str] = Field(default=None, description="Traceback of the error")
    timestamp: str = Field(
        default_factory=lambda: datetime.datetime.now().isoformat(),
        description="Time of the event",
    )


class HealthCheckResponse(BaseModel):
    """Represents a health check response"""

    status: Literal["ok", "error"] = Field(..., description="Server status")
    unhealthy_events: list[UnhealthyEvent] = Field(
        default_factory=list, description="List of unhealthy events"
    )



from collections import deque
from .types import UnhealthyEvent

__all__ = ["manager"]


class HealthManager:
    """Manages the health of the server"""

    UnhealthyEvents: deque[UnhealthyEvent] = deque(
        maxlen=100
    )  # we do not want to memory leak

    def add_unhealthy_event(self, event: UnhealthyEvent) -> None:
        self.UnhealthyEvents.append(event)

    def get_unhealthy_events(self) -> list[UnhealthyEvent]:
        return list(self.UnhealthyEvents)

    def is_healthy(self) -> bool:
        return not any(event.severity == "error" for event in self.UnhealthyEvents)


manager: HealthManager = HealthManager()



import asyncio
from mcp import StdioServerParameters, stdio_client

from mcp_bridge.config import config
from mcp_bridge.mcp_clients.session import McpClientSession
from .AbstractClient import GenericMcpClient
from loguru import logger
import shutil
import os


# Keywords to identify virtual environment variables
venv_keywords = ["CONDA", "VIRTUAL", "PYTHON"]

class StdioClient(GenericMcpClient):
    config: StdioServerParameters

    def __init__(self, name: str, config: StdioServerParameters) -> None:
        super().__init__(name=name)

        # logger.debug(f"initializing settings for {name}: {config.command} {" ".join(config.args)}")

        own_config = config.model_copy(deep=True)

        env = dict(os.environ.copy())

        env = {
            key: value for key, value in env.items()
            if not any(key.startswith(keyword) for keyword in venv_keywords)
        }

        if config.env is not None:
            env.update(config.env)

        own_config.env = env

        command = shutil.which(config.command)
        if command is None:
            logger.error(f"could not find command {config.command}")
            exit(1)

        own_config.command = command

        # this changes the default to ignore
        if "encoding_error_handler" not in config.model_fields_set:
            own_config.encoding_error_handler = "ignore"

        self.config = own_config

    async def _maintain_session(self):
        logger.debug(f"starting maintain session for {self.name}")
        async with stdio_client(self.config) as client:
            logger.debug(f"entered stdio_client context manager for {self.name}")
            assert client[0] is not None, f"missing read stream for {self.name}"
            assert client[1] is not None, f"missing write stream for {self.name}"
            async with McpClientSession(*client) as session:
                logger.debug(f"entered client session context manager for {self.name}")
                await session.initialize()
                logger.debug(f"finished initialise session for {self.name}")
                self.session = session

                try:
                    while True:
                        await asyncio.sleep(10)
                        if config.logging.log_server_pings:
                            logger.debug(f"pinging session for {self.name}")

                        await session.send_ping()

                except Exception as exc:
                    logger.error(f"ping failed for {self.name}: {exc}")
                    self.session = None

        logger.debug(f"exiting session for {self.name}")



from httpx import AsyncClient
from mcp_bridge.config import config

client: AsyncClient = AsyncClient(
    base_url=config.inference_server.base_url,
    headers={"Authorization": f"Bearer {config.inference_server.api_key}", "Content-Type": "application/json"},
    timeout=10000,
)



from typing import Annotated, Literal, Union
from pydantic_settings import BaseSettings, SettingsConfigDict
from pydantic import BaseModel, Field

from mcp.client.stdio import StdioServerParameters


class InferenceServer(BaseModel):
    base_url: str = Field(
        default="http://localhost:11434/v1",
        description="Base URL of the inference server",
    )
    api_key: str = Field(
        default="unauthenticated", description="API key for the inference server"
    )


class Logging(BaseModel):
    log_level: Literal["INFO", "DEBUG"] = Field("INFO", description="default log level")
    log_server_pings: bool = Field(False, description="log server pings")


class SamplingModel(BaseModel):
    model: Annotated[str, Field(description="Name of the sampling model")]

    intelligence: Annotated[float, Field(description="Intelligence of the sampling model")] = 0.5
    cost: Annotated[float, Field(description="Cost of the sampling model")] = 0.5
    speed: Annotated[float, Field(description="Speed of the sampling model")] = 0.5


class Sampling(BaseModel):
    timeout: Annotated[int, Field(description="Timeout for sampling requests")] = 10
    models: Annotated[list[SamplingModel], Field(description="List of sampling models")] = []

class SSEMCPServer(BaseModel):
    # TODO: expand this once I find a good definition for this
    url: str = Field(description="URL of the MCP server")

class DockerMCPServer(BaseModel):
    container_name: str | None = Field(default=None, description="Name of the docker container")
    image: str = Field(description="Image of the docker container")
    args: list[str] = Field(default_factory=list, description="Command line arguments for the docker container")
    env: dict[str, str] = Field(default_factory=dict, description="Environment variables for the docker container")


MCPServer = Annotated[
    Union[StdioServerParameters, SSEMCPServer, DockerMCPServer],
    Field(description="MCP server configuration"),
]


class Network(BaseModel):
    host: str = Field("0.0.0.0", description="Host of the network")
    port: int = Field(8000, description="Port of the network")


class Settings(BaseSettings):
    inference_server: InferenceServer = Field(
        default_factory=lambda: InferenceServer.model_construct(),
        description="Inference server configuration"
    )

    mcp_servers: dict[str, MCPServer] = Field(
        default_factory=dict, description="MCP servers configuration"
    )

    sampling: Sampling = Field(
        default_factory=lambda: Sampling.model_construct(),
        description="sampling config",
    )

    logging: Logging = Field(
        default_factory=lambda: Logging.model_construct(),
        description="logging config",
    )

    network: Network = Field(
        default_factory=lambda: Network.model_construct(),
        description="network config",
    )

    model_config = SettingsConfigDict(
        env_prefix="MCP_BRIDGE__",
        env_file=".env",
        env_file_encoding="utf-8",
        env_nested_delimiter="__",
        cli_parse_args=True,
        cli_avoid_json=True,
    )



from mcp_bridge.config.initial import initial_settings
from mcp_bridge.config.final import Settings
from typing import Any, Callable
from loguru import logger
from pydantic import ValidationError

__all__ = ["config"]

config: Settings = None  # type: ignore

if initial_settings.load_config:
    # import stuff needed to load the config
    from deepmerge import always_merger
    import sys

    configs: list[dict[str, Any]] = []
    load_config: Callable[[str], dict]  # without this mypy will error about param names

    # load the config
    if initial_settings.file is not None:
        logger.info(f"Loading config from {initial_settings.file}")
        from .file import load_config

        configs.append(load_config(initial_settings.file))

    if initial_settings.http_url is not None:
        logger.info(f"Loading config from {initial_settings.http_url}")
        from .http import load_config

        configs.append(load_config(initial_settings.http_url))

    if initial_settings.json is not None:
        logger.info("Loading config from json string")
        configs.append(initial_settings.json)

    # merge the configs
    result: dict = {}
    for cfg in configs:
        always_merger.merge(result, cfg)

    # build the config
    try:
        config = Settings(**result)
    except ValidationError as e:
        logger.error("unable to load a valid configuration")
        for error in e.errors():
            logger.error(f"{error['loc'][0]}: {error['msg']}")
        exit(1)

    if config.logging.log_level != "DEBUG":
        logger.remove()
        logger.add(
            sys.stderr,
            format="{time} {level} {message}",
            level=config.logging.log_level,
            colorize=True,
        )
